# vcr-lib.tcl --
#
#       Library of abstractions for controlling the VCR through the amxd.
#
# Copyright (c) 2000-2002 The Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# A. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# B. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# C. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# these functions must be called from a program with direct access to the
#   amx-lib functions (i.e., on the AMX Daemon server)
#
# this library assumes that the AMX has already been setup with
#   setup-AMX-control

source amx-lib.tcl

# these are defined in amx-lib.tcl
#global g_amxDevices g_amxChannels

proc vcr_doFunction { id func} {
    global g_amxDevices g_amxChannels

    switch -exact -- $id {
	front {
	    set dev $g_amxDevices(frontVCR)
	}
	rack {
	    set dev $g_amxDevices(rackVCR)
	}
	default {
	    return -code error "id must be \[front, rack\]"
	}
    }

    switch -exact -- $func {
	play -
	stop -
	pause -
	ffwd -
	rew -
	rec -
	power {
	    # controls are the same for front and rack VCRs
	    send-AMX-command 10 $dev $g_amxChannels(frontVCR,$func)
	}
	default {
	    return -code error "func must be \[play, stop, pause, ffwd, rew, rec, power\]"
	}
    }
}

