/*
 * encoder-lpc.cc --
 *
 *      LPC Audio encoder
 *
 * Copyright (c) 1995-2002 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * A. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * B. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * C. Neither the names of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /usr/mash/src/repository/mash/mash-1/codec/audio/encoder-lpc.cc,v 1.7 2002/02/03 03:11:42 lim Exp $";
#endif

#include "tclcl.h"
#include "config.h"
#include "encoder.h"
#include "rtp.h"
#include "transmitter.h"
extern "C" {
#include "lpc.h"
}

class LPCEncoder : public EncoderModule {
public:
	LPCEncoder();
protected:
	void encode(u_int32_t ts, const u_int8_t* samples, int cc);
	u_int32_t last_ts_;
};

static class LPCEncoderClass : public TclClass {
public:
	LPCEncoderClass() : Matcher("Module/Encoder/LPC") {}
	TclObject* create(int, const char*const*) {
		return (new LPCEncoder);
	}
} lpc_encoder_class;

LPCEncoder::LPCEncoder()
{
	lpcstate_t dummy;
	lpc_init(&dummy);
}

void LPCEncoder::encode(u_int32_t ts, const u_int8_t* frame, int len)
{
	register int marker = (ts != last_ts_) ? htons(RTP_M) : 0;
	last_ts_ = ts + len;

	pktbuf* pb = pool_->alloc(ts, RTP_PT_LPC);
	rtphdr* rh = (rtphdr*)pb->hdr;
	rh->rh_flags |= marker;
	pb->iov[0].iov_len = sizeof(rtphdr);
	u_int8_t* o = (u_int8_t*)pb->iov[1].iov_base;
	int outlen = 0;
	while (len > 0) {
		lpcparams_t* p = (lpcparams_t*)o;
		lpc_analyze(frame, p);
		HTONS(p->period);
		o += LPCRECSIZE;
		outlen += LPCRECSIZE;
		len -= FRAMESIZE;
		frame += FRAMESIZE;
	}
	pb->iov[1].iov_len = outlen;
	target_->recv(pb);
	return (outlen + sizeof(rtphdr));
}
