/*
 * addition.cc --
 *
 *      A scalar addition effect. It doesn't make anything useful, but it's 
 *      interesting for performance measurements.
 *
 * Copyright (c) 1993-2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * A. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * B. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * C. Neither the names of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* @(#) $Header: /usr/mash/src/repository/mash/mash-1/fx/effects/addition.cc,v 1.1 2002/02/07 04:18:18 chema Exp $ */


#include "addition.h"


AdditionEffect::AdditionEffect(int decimation) 
	: EffectModule(1, decimation)
{
	scale = 0;
}



int AdditionEffect::command(int argc, const char*const* argv)
{
	return (EffectModule::command(argc, argv));
}



void AdditionEffect::trigger_effect (Uncompressed* input)
{
	// get the scale factor
	scale = (scale + 1) % 0xff;

	// step 1: output = input
	output_->byte_copy(input);

	// step 2: output += scale
	int n = output_->w_ * output_->h_;
	// this only affects the first w*h bytes, i.e., the luma
	for (int i = 0; i < n; i++) {
		output_->bp_[i] += scale;
	}

	// deliver the output frame
	deliver_frame();

	return;
}

