# main.tcl --
#
#       This is the main entry point for the cache tool.
#       Tcl source code expansion with tcl-expand.tcl starts here.
#       Tcl execution also starts here.
#
# Copyright (c) 1998-2002 The Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# A. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# B. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# C. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import WebCacheProxy WebCacheControl WebCacheApplication WebCacheApplication/Layer

proc main { args } {

	global cache

	set cacheType [lindex [lindex $args 0] 0]
	set args [lrange [lindex $args 0] 1 end]

	if { $args == "" || "$cacheType" == "-help" } {
		# init will call usage since we pass -help
		puts "cache: \[-layer|-lazy|-base\] \[other-options-listed-below] address"
		set cache [new WebCacheApplication -help]
		delete $cache
		exit 1
	} elseif { "$cacheType" == "-layer" } {
		set cache [new WebCacheApplication/Layer $args]
	} elseif { "$cacheType" == "-lazy" } {
		set cache [new WebCacheApplication/Lazy $args]
	} elseif { "$cacheType" == "-base" }  {
		set cache [new WebCacheApplication $args]
	}

	set proxy [new WebCacheProxy]

	$proxy set cache_ $cache
	$proxy open

	$cache set proxy_ $proxy

	wm withdraw .
}

main $argv

