# front.tcl --
#
#       FIXME: This file needs a description here.
#
# Copyright (c) 1999-2002 The Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# A. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# B. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# C. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import RTPApplication
import GraphComm
import FAgent

Class FX_Forward_Front -superclass RTPApplication

FX_Forward_Front instproc init args {
    $self next fx_forw_front

    $self init_resources

    [$self options] register_option -from_cntrl from_cntrl
    [$self options] register_option -to_cntrl to_cntrl

    eval [$self options] parse_args $args;

    $self instvar from_cntrl_obj_
    $self instvar to_cntrl_obj_

    set from_cntrl_spec [split [$self get_option from_cntrl] "/"]
    set from_addr [lindex $from_cntrl_spec 0];
    set from_port [lindex $from_cntrl_spec 1];
    set from_ttl [lindex $from_cntrl_spec 2];

    set from_cntrl_obj_ [new GraphComm/Front/From $self [$self get_option comm_id] $from_addr $from_port $from_ttl];

    set to_cntrl_spec [split [$self get_option to_cntrl] "/"]
    set to_addr [lindex $to_cntrl_spec 0];
    set to_port [lindex $to_cntrl_spec 1];
    set to_ttl [lindex $to_cntrl_spec 2];

    $self instvar back_host_
    set back_host_ $to_addr;

    set to_cntrl_obj_ [new GraphComm/Front/To $self [$self get_option comm_id] $to_addr $to_port $to_ttl];

    $from_cntrl_obj_ install $to_cntrl_obj_
    $to_cntrl_obj_ install $from_cntrl_obj_
}

FX_Forward_Front instproc back_host {} {
    $self instvar back_host_;

    return $back_host_;
}

FX_Forward_Front instproc set_back_host {host} {
    $self instvar back_host_;

    puts "Setting back host to $host"

    set back_host_ $host;
}

FX_Forward_Front instproc init_resources {} {
    $self add_option network ip
    $self add_option mtu 1024
    $self add_option defaultTTL 32
    $self add_option sessionType rtpv2
    $self add_option maxVideoSessionBW 30000000

    set hname [exec hostname];
    set pid [pid];

    $self add_option comm_id ${hname}.${pid};
}

FX_Forward_Front instproc GenerateNewTunnelSpec {} {
    $self instvar used_ports_;

    set new_port [expr (int(rand() * 10000)*2) + 5000];
    while {[info exists used_ports_($new_port)]} {
	set new_port [expr (int(rand() * 10000)*2) + 5000];
    }
    set used_ports_($new_port) 1;
    return "[exec hostname]/$new_port"
}


Class GraphComm/Front -superclass GraphComm

GraphComm/Front instproc init {app id addr port ttl} {
    $self next $id $addr $port

    $self instvar app_

    set app_ $app;

    $self instvar dest_;

    set dest_ ""
}

GraphComm/Front instproc install {dest} {
    $self instvar dest_;

    set dest_ $dest;
}

GraphComm/Front instproc recv_trigger_command {cmd} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }

    $dest_ send_trigger_command $cmd
}

GraphComm/Front instproc recv_misc {cmd} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }

    $dest_ send_misc $cmd
}

GraphComm/Front instproc recv_debug {data} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }

    $dest_ send_debug $data
}

GraphComm/Front instproc new_input {new_name} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $new_name

    $dest_ create_input $new_name
}

GraphComm/Front instproc new_output {new_name} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $new_name

    $dest_ create_output $new_name
}

GraphComm/Front instproc new_parameter {new_name} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $new_name

    $dest_ create_parameter $new_name
}

GraphComm/Front instproc new_input_attribute {input_name attr_name} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $input_name $attr_name

    $dest_ create_input_attr $input_name $attr_name
}

GraphComm/Front instproc new_output_attribute {output_name attr_name} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $output_name $attr_name

    $dest_ create_output_attr $output_name $attr_name
}

GraphComm/Front instproc new_parameter_attribute {parameter_name attr_name} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $parameter_name $attr_name

    $dest_ create_parameter_attr $parameter_name $attr_name
}

GraphComm/Front instproc update_input_attr_value {input_name attr_name value} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $input_name $attr_name $value

    $dest_ set_input_attr $input_name $attr_name $value
}

GraphComm/Front instproc update_output_attr_value {output_name attr_name value} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $output_name $attr_name $value

    $dest_ set_output_attr $output_name $attr_name $value
}

GraphComm/Front instproc update_parameter_attr_value {parameter_name attr_name value} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }
    $self next $parameter_name $attr_name $value

    $dest_ set_parameter_attr $parameter_name $attr_name $value
}

GraphComm/Front instproc set_input_attr {input_name attr_name value} {
    $self create_input $input_name
    $self create_input_attr $input_name $attr_name
    $self next $input_name $attr_name $value
}

GraphComm/Front instproc set_output_attr {output_name attr_name value} {
    $self create_output $output_name
    $self create_output_attr $output_name $attr_name
    $self next $output_name $attr_name $value
}

GraphComm/Front instproc set_parameter_attr {parameter_name attr_name value} {
    $self create_parameter $parameter_name
    $self create_parameter_attr $parameter_name $attr_name
    $self next $parameter_name $attr_name $value
}

Class GraphComm/Front/From -superclass GraphComm/Front

GraphComm/Front/From instproc update_input_attr_value {input_name attr_name value} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }

    $self instvar app_

    if {[$app_ back_host] == ""} {
	return;
    }

    if {$attr_name == "spec"} {
	set split_spec [split $value "/"];
	set addr [lindex $split_spec 0];
	set port [lindex $split_spec 1];
	set srcid [lindex $split_spec 2];

	$self instvar tunnels_
	if {[info exists tunnels_($addr,$port)]} {
	    set fagent $tunnels_($addr,$port);
	    set tunnel_spec $tunnel_specs_($addr,$port);
	} else {
	    set tunnel_spec [$app_ GenerateNewTunnelSpec];
	    set tunnel_port [lindex [split $tunnel_spec "/"] 1]
	    set fagent [new FAgent $app_ "${addr}/${port}" [$app_ back_host]/$tunnel_port];
	    set tunnels_($addr,$port) $fagent;
	    set tunnel_specs_($addr,$port) $tunnel_spec;
	}
	if {$srcid != "*"} {
	    $fagent install_src_callback $srcid "$self complete_input_setup $srcid $fagent $input_name $tunnel_spec"
	    return;
	} else {
	    set value "${tunnel_spec}/*";
	}
    }
    $self next $input_name $attr_name $value;
}

GraphComm/Front/From instproc complete_input_setup {srcid fagent input_name tunnel_spec} {
    set new_id [$fagent translate_srcid $srcid];

    $self instvar dest_;

    $dest_ set_input_attr $input_name spec "${tunnel_spec}/${new_id}"
}

GraphComm/Front/From instproc update_output_attr_value {output_name attr_name value} {
    $self instvar dest_;

    if {$dest_ == ""} {
	return;
    }

    $self instvar app_

    if {$attr_name == "spec"} {
	$self instvar out_tunnels_
	$self instvar out_tunnel_specs_

	if {[info exists out_tunnels_($value)]} {
	    set fagent $out_tunnels_($value);
	    set tunnel_spec $out_tunnel_specs_($value);
	} else {
	    set tunnel_spec [$app_ GenerateNewTunnelSpec];
	    set fagent [new FAgent $app_ $tunnel_spec $value];
	    set out_tunnels_($value) $fagent;
	    set out_tunnel_specs_($value) $tunnel_spec;
	}
	set value $tunnel_spec
    }
    $self next $output_name $attr_name $value;
}


Class GraphComm/Front/To -superclass GraphComm/Front

GraphComm/Front/To instproc recv_misc {cmd} {
    if {[lindex $cmd 0] == "set_back_host"} {
	$self instvar app_;

	$app_ set_back_host [lindex $cmd 1];
    } else {
	$self next $cmd;
    }
}

set app [new FX_Forward_Front $argv];
